import json

import requests
from jinja2 import Environment, PackageLoader, select_autoescape, FileSystemLoader


class SpackPackage:
    content: dict = {}
    versions: list = []
    url: str = ""
    homepage: str = ""
    maintainers: list = []
    package_name: str = ''
    summary: str = ''

    def __init__(self, content):
        self.content = self._download_missing_information(content)
        self._url()

        self._versions_formatter()
        self.version_builder()

        self.package_name_builder()
        self.summary_builder()
        self.homepage_builder()

    @property
    def get_versions(self):
        self.versions = []
        self._versions_formatter()
        self.version_builder()
        return self.versions

    def _download_missing_information(self, content):
        if isinstance(content, str):
            if content.startswith('py-'):
                content = content.replace("py-", '')
            status = requests.get(f'https://pypi.org/pypi/{content}/json')
            if status.status_code != 200:
                print("Package doesnt not exist at PyPi.org")
                raise
            return json.loads(status.content)
        else:
            return content

    def _url(self):
        def _generate_pypi_uri(filename_with_version):
            return f'https://pypi.io/packages/source/{filename_with_version[0]}/' \
                   f'{filename_with_version.split("-")[0]}/{filename_with_version}'

        latest_filename: str = ''
        for url in self.content["urls"]:
            if url["url"].endswith(".tar.gz"):
                latest_filename = url["url"].split('/')[-1]
                break
        self.url = _generate_pypi_uri(latest_filename)

        if not self.url:
            print("Source package was not found")

    def _homepage(self):
        self.homepage = self.content['home_page']

    def _versions_formatter(self):
        for version, version_content in self.content['releases'].items():
            only_sdist_version = [x for x in version_content if
                                  (x["url"].endswith(".tar.gz") and x["packagetype"] == "sdist")]
            if not only_sdist_version:
                continue
            for proper_version in only_sdist_version:
                try:
                    # TODO Not support .whl packages yet. Only source code in .tar.gz
                    if not proper_version['filename'].endswith('.tar.gz'):
                        continue
                    self.versions.append(
                        (str(version), str(proper_version['digests']['sha256']))
                    )
                except IndexError:
                    continue

    def version_builder(self) -> str:
        raw_versions: str = ''
        for v in self.versions:
            raw_versions += f'version(\"{str(v[0])}\", sha256=\"{str(v[1])}\")\n'
        return raw_versions

    def package_name_builder(self) -> str:
        name = self.content['info']['name']
        if name[0].islower():
            name = name[0].upper() + name[1:]

        l_name = list(name)
        for i in [index for index, element in enumerate(l_name) if element == '-']:
            try:
                if l_name[i + 1].islower():
                    l_name[i + 1] = l_name[i + 1].upper()
            except:
                pass
            finally:
                name = ''.join(l_name).replace('-', '')

        if not name.startswith("Py"):
            name = "Py" + name
        self.package_name = name
        return name

    def summary_builder(self):
        self.summary = self.content['info']['summary']

    def homepage_builder(self):
        self.homepage = self.content['info']['project_urls']['Homepage']

    def _generate_data(self) -> dict:
        return {
            "package_name": self.package_name,
            "summary": self.summary,
            "homepage": self.homepage,
            "url": self.url,
            "versions": self.versions,
            "depends_on_setuptools": False
        }

    def generate_file(self):
        env = Environment(loader=PackageLoader('pip2spack', 'templates'),
                          autoescape=select_autoescape(['j2']))
        template = env.get_template('package.py')
        return template.render(**self._generate_data())

    def generate_custom_file(self, abspath: str, versions):
        env = Environment(loader=FileSystemLoader(abspath))
        template = env.get_template('package.py')
        return template.render({"versions": versions})



from unittest import TestCase

from pip2spack.spack_package import SpackPackage
from pip2spack.test_data import test_data as ready_packages


class TestSpackPackage(TestCase):
    def setUp(self) -> None:
        self.spack = SpackPackage(ready_packages["jsl"])

    def test_package_name_builder(self):
        package_name = self.spack.package_name_builder()
        self.assertEqual("PyJsl", package_name)

    def test_package_name_builder_replace_dash(self):
        self.spack.content['info']['name'] = "jsl-up"
        package_name = self.spack.package_name_builder()
        self.assertEqual("PyJslUp", package_name)

    def test_generator_is_not_null(self):
        return_message = self.spack._generate_data()
        for key, value in return_message.items():
            if key in ("depends_on_setuptools", ):
                continue
            self.assertTrue(value, msg=f'{key} is not true')



# Copyright 2013-2020 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

# Package automatically generated using 'pip2spack' converter

from spack import *


class {{ package_name }}(PythonPackage):
    """
    {{ summary }}
    """

    homepage = "{{ homepage }}"
    url      = "{{ url }}"
    {% for v in versions|reverse %}
    version('{{ v[0] }}', sha256='{{ v[1] }}'){% endfor %}


    # depends_on('python@2.7:2.8,3.4:', type=('build', 'run'))
    {% if depends_on_setuptools %}
    depends_on('py-setuptools', type='build')
    {% else %}
    # depends_on('py-setuptools', type='build')
    {% endif %}

    def build_args(self, spec, prefix):
        # FIXME: Add arguments other than --prefix
        # FIXME: If not needed delete this function
        args = []
        return args



BUILDIN_REPOSITORY_REL_PATH = "var/spack/repos/builtin/packages"

MARKER_VERSION = "{% pip2spack_version_marker %}"
MARKER_VERSION_TEMPLATE = """{% for v in versions|reverse %}
version('{{ v[0] }}', sha256='{{ v[1] }}'){% endfor %}"""


def marker_version_template(space: int = 0):
    spaces = ' ' * space
    return "{% for v in versions|reverse %}\n" + spaces + "version('{{ v[0] }}', sha256='{{ v[1] }}'){% endfor %}"



import argparse
import json
import os

import requests

from pip2spack.spack_package import SpackPackage
from spack import Spack, Package


def main():
    parser = argparse.ArgumentParser(description='Create a spack package base on:')
    parser.add_argument('-u', '--update', action='store_true')
    parser.add_argument('name', type=str, nargs='+', help='Package name on the pypi.org')

    args = parser.parse_args()

    ready_packages = validate_pip_package_exists(args.name)
    show_packages_for_process(ready_packages, args.name)

    if not args.update:
        generate_packages(ready_packages)
    else:
        update_packages(ready_packages)


def update_packages(ready_packages):
    for p in ready_packages:
        package = Package(package_name=p)
        package.replace_structure_with_marker()
        package.generate_modded_package()


def generate_packages(ready_packages):
    for key, value in ready_packages.items():
        print(f"INFO:: Generating package for the {key}")
        s_package = SpackPackage(value)
        generated_file = s_package.generate_file()

        create_directory(key)
        created_package_uri = create_package(key, generated_file)
        print(f"_OK_:: Package for {key} was generated\n\t"
              f":: If \'spack install py-{key}\' return any problems, try to uncomment:\n\t"
              f":: \tdepends_on('py-setuptools', type='build')\n\t"
              f":: at {created_package_uri}")


def validate_pip_package_exists(potential_packages):
    if not potential_packages:
        return

    packages_status = {}
    for package in potential_packages:
        status = requests.get(f'https://pypi.org/pypi/{package}/json')
        if status.status_code != 200:
            continue
        packages_status[package] = json.loads(status.content)
    return packages_status


def show_packages_for_process(package_ready, all_packages):
    """Display at the terminal a information about arability of provided package names"""
    print("Packages ready to convert:")
    for key, _ in package_ready.items():
        print(f"\t{key}")

    packages_not_found = [x for x in all_packages if x not in package_ready]
    if not packages_not_found:
        print()
        return

    print("\nPackages not found:")
    for p in packages_not_found:
        print(f'\t{p}')
    else:
        print()


def get_spack_repository() -> str:
    """Return a absolute path about spack repository"""
    if not os.getenv("SPACK_ROOT", False):
        print('Please provide a information about SPACK_ROOT')
        exit(1)

    repository_suburi = os.path.relpath('var/spack/repos/builtin/packages/')
    repository_full_path = os.path.join(os.getenv("SPACK_ROOT"), repository_suburi)

    if not os.path.exists(repository_full_path):
        print('Spack repository does not exists')
        exit(1)
    return repository_full_path


def create_directory(package_name):
    """Create a package directory for the provided package name"""
    if not package_name.startswith('py-'):
        package_name = 'py-' + package_name
    target_path = os.path.join(get_spack_repository(),
                               package_name)
    if os.path.exists(target_path):
        return False
    try:
        os.makedirs(target_path)
    except FileExistsError:
        pass
    finally:
        return os.path.exists(target_path)


def create_package(name, raw) -> str:
    if not name.startswith('py-'):
        name = 'py-' + name
    target_path = os.path.join(get_spack_repository(),
                               name)
    with open(os.path.join(target_path, 'package.py'), 'w') as f:
        f.write(raw)

    return os.path.join(target_path, 'package.py')


if __name__ == "__main__":
    main()



import os
from dataclasses import dataclass
from os import getenv, listdir
from os.path import exists, join, relpath, abspath
from typing import AnyStr

import constants
from constants import BUILDIN_REPOSITORY_REL_PATH
from spack_package import SpackPackage


@dataclass
class Spack:
    _spack_root: str = getenv("SPACK_ROOT", None)
    _builtin_repository: str = f'{join(_spack_root, relpath(BUILDIN_REPOSITORY_REL_PATH))}' if _spack_root else ""

    @property
    def directory(self):
        return self._spack_root

    @directory.setter
    def directory(self, custom_spack_root):
        if not exists(custom_spack_root):
            raise
        self._spack_root = custom_spack_root

    @property
    def builtin_repository(self):
        return self._builtin_repository

    def list_all_packages(self):
        return listdir(self.builtin_repository)

    def list_py_packages(self):
        return [x for x in self.list_all_packages() if x.startswith("py-")]

    def exists(self, package_name: str, python_only: bool = True) -> bool:
        if python_only:
            return package_name in self.list_py_packages()
        else:
            return package_name in self.list_all_packages()

    def get_package_path(self, package_name: str) -> AnyStr:
        """Return absolute path to the package.py if exists in builtin repository"""
        if not package_name.startswith('py-'):
            package_name = "py-" + package_name
        package_py_path: str = join(*(f"{self.builtin_repository}", package_name, "package.py"))
        if not self.exists(package_name) or not exists(package_py_path):
            # TODO create a new one
            print(f"Package {package_name} does not exist in the builtin repository")
            raise
        return abspath(package_py_path)


@dataclass
class Package:
    package_name: AnyStr
    package_path: AnyStr = None
    _raw_package: AnyStr = None

    versions = []

    def __post_init__(self):
        s = Spack()
        self.package_path = s.get_package_path(package_name=self.package_name)
        self._open()

    def _open(self):
        with open(self.package_path, 'r') as package_py:
            self._raw_package = package_py.read()

    def replace_structure_with_marker(self):
        self.modificated_package = self._raw_package.split('\n')
        if not isinstance(self.modificated_package, list):
            raise

        rows_with_version = []
        for index, content in enumerate(self.modificated_package):
            if "version(" in content:
                rows_with_version.append((index, content.lstrip().rstrip()))

        rows_with_version.reverse()  # start removing a empty rows from bottom file
        for index, row in enumerate(rows_with_version):
            if index == 0:
                self.modificated_package[row[0]] = self.modificated_package[row[0]].replace(row[1],
                                                                                            constants.MARKER_VERSION)
            else:
                del self.modificated_package[row[0]]

        self.replace_marker_with_template()

    def replace_marker_with_template(self):
        rows_with_version = []
        for index, content in enumerate(self.modificated_package):
            if constants.MARKER_VERSION in content:
                rows_with_version.append(index)
        rows_with_version.reverse()  # start removing a empty rows from bottom file
        for index, row in enumerate(rows_with_version):
            if index == 0:
                spaces = self._count_trailing_spaces(self.modificated_package[row])
                self.modificated_package[row] = \
                    self.modificated_package[row].replace(constants.MARKER_VERSION,
                                                          constants.marker_version_template(spaces))
            else:
                break
        self.generate_modded_package(save=True)
        if not self.save(self.update_newest_versions(self.package_name)):
            print(f"ERRR:: Something went wrong, package {self.package_name} was not updated")
        else:
            print(f" OK :: {self.package_name} :: builtin package was updated at {self.package_path}")

    def update_newest_versions(self, package_name):
        sp = SpackPackage(package_name)
        file = sp.generate_custom_file(os.path.dirname(os.path.realpath(self.package_path)), sp.get_versions)
        return file

    def save(self, new_file):
        with open(self.package_path, 'w') as package:
            package.write(new_file)
            return True

    def generate_modded_package(self, save: bool = False):
        if not save:
            return '\n'.join(self.modificated_package).lstrip()
        self.save('\n'.join(self.modificated_package).lstrip())

    @staticmethod
    def available_markers(marker_name):
        return {"version": constants.MARKER_VERSION}.get(marker_name, None)

    @staticmethod
    def _count_trailing_spaces(sentence) -> int:
        return len(sentence) - len(sentence.lstrip())


if __name__ == "__main__":
    # TODO generate tests base on this queries
    package = Package(package_name="py-codecov")
    print(package.replace_structure_with_marker())
    print(package.generate_modded_package())
